#
# Copyright 2009 Canonical Ltd.
#
# Written by:
#     Gustavo Niemeyer <gustavo.niemeyer@canonical.com>
#     Sidnei da Silva <sidnei.da.silva@canonical.com>
#
# This file is part of the Image Store Proxy.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
import hashlib
import time
import os

from urllib import  quote

from imagestore.lib.fetch import fetch, FetchError
from imagestore.lib.service import (
    ServiceTask, ServiceError, ThreadedService, taskHandlerInThread)


CHUNK_SIZE = 1<<16


def verifySHA256(filePath, checksum):
    sha256 = hashlib.sha256()
    handle = open(filePath, 'r')
    try:
        while True:
            data = handle.read(CHUNK_SIZE)
            if not data:
                break
            sha256.update(data)
        return sha256.hexdigest() == checksum
    finally:
        handle.close()


class DownloadServiceError(ServiceError):
    pass


class Progress(object):

    def __init__(self):
        self._abort = False
        self._active = False
        self._partialSize = 0
        self._expectedSize = 0
        self._currentSize = 0
        self._totalSize = 0
        self._lastLog = 0


    def cancel(self):
        self._abort = True

    def setPartialSize(self, partialSize):
        """Set to the size of a partially downloaded file if any.
        """
        self._partialSize = partialSize

    def setExpectedSize(self, size):
        """Set the expected size of the file to be downloaded.
        """
        self._expectedSize = size

    def getCurrentSize(self):
        """Get the current size of the file being downloaded.
        """
        return self._currentSize

    def getTotalSize(self):
        """Get the total size of the file being downloaded.
        """
        return self._totalSize

    def isActive(self):
        return self._active

    def __call__(self, downTotal, downCurrent, upTotal, upCurrent):
        current = None
        total = None

        if not downTotal:
            # XXX This branch is untested.
            if self._expectedSize and downCurrent:
                current, total = self._partialSize + downCurrent, self._expectedSize
            else:
                # XXX This branch is untested as well.
                pass
        else:
            current, total = (self._partialSize + downCurrent,
                              self._partialSize + downTotal)

        if current is not None and total is not None:
            self._currentSize = current
            self._totalSize = total

            now = time.time()
            if self._lastLog < now - 5:
                self._lastLog = now
                logging.debug("Download in progress (current=%d, total=%d)" %
                              (self._currentSize, self._totalSize))

        if self._abort:
            # Returning a non-zero integer value cancels the download.
            return 1


class DownloadServiceTask(ServiceTask):
    pass


class DownloadFileTask(DownloadServiceTask):

    def __init__(self, url, size, sha256):
        self.url = url
        self.size = size
        self.sha256 = sha256
        self.progress = Progress()


class DownloadService(ThreadedService):

    def __init__(self, reactor, basePath):
        self._basePath = basePath
        ThreadedService.__init__(self, reactor)

    @taskHandlerInThread(DownloadFileTask)
    def _downloadFile(self, task):
        logging.info("Download starting for %s" % (task.url,))

        def progress(partialSize, totalSize):
            task.progress.setPartialSize(partialSize)
            task.progress.setExpectedSize(totalSize)
            return task.progress

        localPath = os.path.join(self._basePath, quote(task.url, safe=""))
        try:
            fetch(task.url, size=task.size,
                  resume=True, local_path=localPath,
                  progress=progress)
        except FetchError, e:
            logging.error("Download failed: %s" % (e,))
            raise DownloadServiceError(str(e))
        else:
            logging.info("Download finished.")

        logging.debug("Verifying checksum of downloaded file.")
        if not verifySHA256(localPath, task.sha256):
            message = "Checksum mismatch on downloaded file (expected %s)" \
                      % (task.sha256,)
            logging.error(message)
            raise DownloadServiceError(message)
        logging.info("Checksum of downloaded file matches.")
        return localPath
