#
# Copyright 2009 Canonical Ltd.
#
# Written by:
#     Gustavo Niemeyer <gustavo.niemeyer@canonical.com>
#     Sidnei da Silva <sidnei.da.silva@canonical.com>
#
# This file is part of the Image Store Proxy.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import base64

from urllib import quote

from imagestore.lib.service import (
    ServiceTask, ServiceError, Service, taskHandlerInThread)
from imagestore.lib.twistedutil import mergeDeferreds

from imagestore.model import ImageState, ImageStoreResponse, ImageRegistration
from imagestore.downloadservice import DownloadFileTask, DownloadServiceError
from imagestore.eucaservice import BundleAndUploadImageTask, EucaServiceError
from imagestore.storageservice import (
    GetStoredImagesTask, GetStoredImageStatesTask, StorageServiceError,
    SetImageRegistrationTask)


class InstallServiceError(ServiceError):
    pass


class InstallServiceTask(ServiceTask):
    pass


class GetStatesTask(InstallServiceTask):

    def __init__(self, uris):
        self.uris = uris


class GetInstalledImagesTask(InstallServiceTask):
    pass


class InstallImageTask(InstallServiceTask):

    def __init__(self, image):
        self.image = image


class InstallService(Service):

    def __init__(self, serviceHub):
        self._serviceHub = serviceHub
        self._downloads = {}
        self._downloading = set()
        self._installing = set()
        Service.__init__(self)

    @taskHandlerInThread(InstallImageTask)
    def _installImage(self, task):
        deferreds = []

        imageUri = task.image["uri"]
        imageRegistration = ImageRegistration(task.image)

        imageFiles = {}
        for imageFile in task.image["files"]:
            imageFiles[imageFile["kind"]] = imageFile.copy()

        downloads = self._downloads[imageUri] = []
        for kind in ("kernel", "ramdisk", "image"):
            imageFile = imageFiles[kind]

            if kind == "kernel":
                imagePart = imageRegistration.eki
            elif kind == "ramdisk":
                imagePart = imageRegistration.eri
            elif kind == "image":
                imagePart = imageRegistration.emi

            imagePart.size = imageFile["size-in-bytes"]
            imagePart.sha256 = imageFile["sha256"]
            imagePart.kind = kind

            downloadTask = DownloadFileTask(imageFile["url"],
                                    imageFile["size-in-bytes"],
                                    imageFile["sha256"])

            downloads.append(downloadTask)
            deferred = self._serviceHub.addTask(downloadTask)
            deferred.addCallback(
                lambda result, imagePart=imagePart: (imagePart, result))
            deferreds.append(deferred)

        def installFinishedCallback(result, imageUri):
            self._installing.remove(imageUri)
            return result

        def installFailedErrback(failure, imageUri):
            for downloadTask in self._downloads.get(imageUri, ()):
                downloadTask.progress.cancel()
            self._downloads.pop(imageUri, None)
            self._downloading.discard(imageUri)
            self._installing.discard(imageUri)
            return failure

        self._downloading.add(imageUri)
        deferred = mergeDeferreds(deferreds)
        deferred.addCallback(self._bundleImage, imageRegistration)
        deferred.addCallback(installFinishedCallback, imageUri)
        deferred.addErrback(installFailedErrback, imageUri)

        return deferred

    def _bundleImage(self, result, imageRegistration):
        for imagePart, filePath in result:
            imagePart.path = filePath

        # Mark the image as installing and clear the internal state
        # that's keeping track of downloads.
        self._installing.add(imageRegistration.image["uri"])
        self._downloading.remove(imageRegistration.image["uri"])
        del self._downloads[imageRegistration.image["uri"]]

        def bundleAndUploadCallback(imageRegistration):
             deferred = self._serviceHub.addTask(
                 SetImageRegistrationTask(imageRegistration))
             deferred.addCallback(
                 lambda result, imageRegistration=imageRegistration:
                 imageRegistration)
             return deferred
        deferred = self._serviceHub.addTask(
            BundleAndUploadImageTask(imageRegistration))
        deferred.addCallback(bundleAndUploadCallback)
        return deferred

    @taskHandlerInThread(GetInstalledImagesTask)
    def _getInstalledImages(self, task):
        # XXX Untested
        return []

    @taskHandlerInThread(GetStatesTask)
    def _getStates(self, task):
        # XXX Untested
        def getStoredImageStatesCallback(states):
            for state in states:
                if state["image-uri"] in self._downloading:
                    downloads = self._downloads.get(state["image-uri"])
                    for downloadTask in downloads:
                        # XXX Compute progress
                        pass
                    state["status"] = "downloading"
                elif state["image-uri"] in self._installing:
                    state["status"] = "installing"
                else:
                    # XXX Untested. Need to add actions here.
                    state["actions"] = {
                        "install": (
                            "http://localhost:52780/api/images/%s/install" %
                            base64.urlsafe_b64encode(state["image-uri"])),
                        }
            return states

        deferred = self._serviceHub.addTask(GetStoredImageStatesTask(task.uris))
        deferred.addCallback(getStoredImageStatesCallback)
        return deferred
