#
# Copyright 2009 Canonical Ltd.
#
# Written by:
#     Gustavo Niemeyer <gustavo.niemeyer@canonical.com>
#     Sidnei da Silva <sidnei.da.silva@canonical.com>
#
# This file is part of the Image Store Proxy.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging

from twisted.internet.defer import succeed

from imagestore.lib.service import ServiceError

from imagestore.installservice import (
    GetStatesTask, GetInstalledImagesTask, InstallImageTask)
from imagestore.model import ImageStoreResponse
from imagestore.proxyservice import GetDashboardTask, GetUpgradesTask
from imagestore.storageservice import (
    SetUpgradeImagesTask, StoreImagesTask, GetStoredImagesTask)
from imagestore.eucaservice import GetEucaInfo


CACHED_IMAGES_THRESHOLD = 30


def gotServiceError(failure):
    failure.trap(ServiceError)
    return ImageStoreResponse({"error-message": failure.value.message})


class Presenter(object):

    def __init__(self, serviceHub):
        self._serviceHub = serviceHub
        self._getStateArgs = object()

    def getDashboard(self):
        logging.debug("Retrieving dashboard.")
        def getUpgrades(uris):
            if uris is not None and uris:
                return self._serviceHub.addTask(
                    GetUpgradesTask(uris))

        def resetStoreAndGetDashboard(upgradeImages):
            if upgradeImages is not None and upgradeImages:
                self._serviceHub.addTask(
                    SetUpgradeImagesTask(upgradeImages))
                return upgradeImages

        def getDashboardWithUpgrades(upgradeImages):
            upgrade_uris = []
            if upgradeImages is not None and upgradeImages:
                for image in upgradeImages:
                    upgrade_uris.append(image["uri"])
            return self._serviceHub.addTask(
                GetDashboardTask(upgrade_uris))

        def storeImages(imageStoreResponse):
            deferred = self._serviceHub.addTask(
                StoreImagesTask(imageStoreResponse["images"]))
            deferred.addCallback(lambda result: imageStoreResponse)
            return deferred

        def getStatesWithImageStoreResponse(imageStoreResponse):
            uris = []
            for image in imageStoreResponse.get("images", ()):
                uris.append(image["uri"])

            def updateStates(states):
                imageStoreResponse["states"] = states
                return imageStoreResponse

            deferred = self._serviceHub.addTask(
                GetStatesTask(uris))
            deferred.addCallback(updateStates)
            return deferred

        deferred = self._serviceHub.addTask(
            GetInstalledImagesTask())
        deferred.addCallback(getUpgrades)
        deferred.addCallback(resetStoreAndGetDashboard)

        deferred.addCallback(getDashboardWithUpgrades)
        deferred.addCallback(storeImages)
        deferred.addCallback(getStatesWithImageStoreResponse)
        deferred.addErrback(gotServiceError)
        return deferred

    def getStates(self, imageUris):
        if imageUris != self._getStateArgs:
            self._getStateArgs = imageUris
            logging.debug("Retrieving states for URIs: %r "
                          "(won't repeat while URIs don't change)"
                          % (imageUris,))
        if not imageUris:
            return succeed([])
        return self._serviceHub.addTask(GetStatesTask(imageUris))

    def search(self, query):
        raise NotImplementedError()

    def startInstall(self, imageUri):
        logging.debug("Starting install of image %s" % (imageUri,))
        deferred = self._serviceHub.addTask(GetStoredImagesTask([imageUri]))
        def installImage(result):
            return self._serviceHub.addTask(InstallImageTask(result[0]))
        deferred.addCallback(installImage)
        return deferred

    def getSecretKey(self, clientId):
        # XXX Untested
        logging.debug("Presenter asking EucaService for credentials "
                      "from admin.")
        deferred = self._serviceHub.addTask(
            GetEucaInfo("admin"))
        deferred.addCallback(lambda result: result.secretKey)
        def logSuccess(result):
            logging.debug("Presenter got EucaInfo and returned it.")
            return result
        def logFailure(result):
            logging.debug("Presenter got an error back from the "
                          "GetEucaInfoTask.")
            return result
        deferred.addCallbacks(logSuccess, logFailure)
        return deferred

    def cancelChange(self, imageUri):
        raise NotImplementedError()

    def clearError(self, imageUri):
        raise NotImplementedError()
