#
# Copyright 2009 Canonical Ltd.
#
# Written by:
#     Gustavo Niemeyer <gustavo.niemeyer@canonical.com>
#     Sidnei da Silva <sidnei.da.silva@canonical.com>
#
# This file is part of the Image Store Proxy.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os

from urllib import quote
from twisted.internet import reactor

from imagestore.lib.tests.mocker import ARGS, KWARGS
from imagestore.lib.fetch import fetch, PyCurlError, HTTPCodeError
from imagestore.lib.twistedutil import mergeDeferreds

from imagestore.tests.helpers import ServiceTestCase
from imagestore.downloadservice import (
    DownloadService, DownloadServiceError, DownloadFileTask)

from imagestore.tests.helpers import ServiceTestCase


class DownloadServiceTest(ServiceTestCase):

    def setUp(self):
        self.basePath = self.makeDir()
        self.service = DownloadService(reactor, self.basePath)

    def testDownloadFile(self):
        image1 = self.createImage(1, withFiles=True)

        tasks = {}
        deferreds = {}
        for imageFile in image1["files"]:
            kind = imageFile["kind"]
            tasks[kind] = task = DownloadFileTask(
                imageFile["url"], imageFile["size-in-bytes"],
                imageFile["sha256"])
            deferreds[kind] = self.service.addTask(task)

        def callbackKernel(result):
            expectedPath = os.path.join(self.basePath,
                                        quote(tasks["kernel"].url, safe=""))
            self.assertEquals(expectedPath, result)
            self.assertEquals(os.path.getsize(expectedPath),
                              tasks["kernel"].size)

        def callbackRAMDisk(result):
            expectedPath = os.path.join(self.basePath,
                                        quote(tasks["ramdisk"].url, safe=""))
            self.assertEquals(expectedPath, result)
            self.assertEquals(os.path.getsize(expectedPath),
                              tasks["ramdisk"].size)

        def callbackImage(result):
            expectedPath = os.path.join(self.basePath,
                                        quote(tasks["image"].url, safe=""))
            self.assertEquals(expectedPath, result)
            self.assertEquals(os.path.getsize(expectedPath),
                              tasks["image"].size)

        deferreds["kernel"].addCallback(callbackKernel)
        deferreds["ramdisk"].addCallback(callbackRAMDisk)
        deferreds["image"].addCallback(callbackImage)
        return self.runServicesAndWaitForDeferred(
            [self.service], mergeDeferreds(deferreds.values()))

    def testDownloadFileFailsChecksum(self):
        image1 = self.createImage(1, withFiles=True)

        tasks = {}
        deferreds = {}
        byKind = {}
        for imageFile in image1["files"]:
            kind = imageFile["kind"]
            byKind[kind] = imageFile

        # In order to force a failed checksum verification, we will
        # force downloading a different url, which will have a
        # different checksum.
        kernelFile = byKind["kernel"]
        imageFile = byKind["image"]
        tasks["kernel"] = task = DownloadFileTask(
            imageFile["url"], imageFile["size-in-bytes"],
            kernelFile["sha256"])
        deferreds["kernel"] = self.service.addTask(task)

        def callbackKernel(result):
            self.fail("Should have raised an exception")

        def errbackKernel(failure):
            failure.trap(DownloadServiceError)
            self.assertStartsWith(failure.value.message,
                                  "Checksum mismatch on downloaded file "
                                  "(expected b90aa07301e80d19137e9f69c5a")

        deferreds["kernel"].addCallback(callbackKernel)
        deferreds["kernel"].addErrback(errbackKernel)
        return self.runServicesAndWaitForDeferred(
            [self.service], mergeDeferreds(deferreds.values()))
