#
# Copyright 2009 Canonical Ltd.
#
# Written by:
#     Gustavo Niemeyer <gustavo.niemeyer@canonical.com>
#     Sidnei da Silva <sidnei.da.silva@canonical.com>
#
# This file is part of the Image Store Proxy.
#
# This program is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License version 3, as published 
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranties of 
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along 
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os

from urllib import quote
from twisted.internet.defer import Deferred

from imagestore.lib.tests.mocker import ARGS, KWARGS
from imagestore.lib.fetch import fetch, PyCurlError, HTTPCodeError
from imagestore.lib.twistedutil import mergeDeferreds
from imagestore.lib.service import (
        Service, ThreadedService, taskHandlerInThread, ServiceHub)

from imagestore.model import ImageState, ImageRegistration
from imagestore.tests.helpers import ServiceTestCase
from imagestore.downloadservice import (
    DownloadService, DownloadServiceError, DownloadFileTask)
from imagestore.eucaservice import (
    EucaService, EucaServiceError, BundleAndUploadImageTask)
from imagestore.installservice import (
    GetStatesTask, InstallService, InstallServiceError, InstallImageTask)
from imagestore.storageservice import (
    GetStoredImagesTask, GetStoredImageStatesTask, StorageServiceError,
    SetImageRegistrationTask)

from imagestore.tests.helpers import ServiceTestCase


class FakeEucaService(Service):

    def __init__(self, imageRegistration):
        self.tasks = []
        self._imageRegistration = imageRegistration
        Service.__init__(self)

    @taskHandlerInThread(BundleAndUploadImageTask)
    def _bundleAndUploadImage(self, task):
        self.tasks.append(task)
        if callable(self._imageRegistration):
            return self._imageRegistration()
        return self._imageRegistration


class FakeStorageService(Service):

    def __init__(self, storedImages, storedImageStates,
                 preImageRegistration=None):
        self.tasks = []
        self._storedImages = storedImages
        self._storedImageStates = storedImageStates
        self._imageRegistration = None
        self._preImageRegistration = preImageRegistration
        Service.__init__(self)

    @taskHandlerInThread(SetImageRegistrationTask)
    def _setImageRegistration(self, task):
        self.tasks.append(task)
        if self._preImageRegistration is not None:
            if callable(self._preImageRegistration):
                self._preImageRegistration()
        self._imageRegistration = task.imageRegistration

    @taskHandlerInThread(GetStoredImagesTask)
    def _getStoredImages(self, task):
        self.tasks.append(task)
        if callable(self._storedImages):
            return self._storedImages()
        return self._storedImages

    @taskHandlerInThread(GetStoredImageStatesTask)
    def _getStoredImageStates(self, task):
        self.tasks.append(task)
        if callable(self._storedImageStates):
            return self._storedImageStates(self._imageRegistration)
        return self._storedImageStates


class FakeDownloadService(Service):

    def __init__(self, downloadFilePaths):
        self.tasks = []
        self._downloadFilePaths = downloadFilePaths
        Service.__init__(self)

    @taskHandlerInThread(DownloadFileTask)
    def _downloadFile(self, task):
        self.tasks.append(task)
        downloadFilePath = self._downloadFilePaths[task.url]
        if callable(downloadFilePath):
            return downloadFilePath()
        return downloadFilePath


class InstallServiceTest(ServiceTestCase):


    def setUp(self):
        self.basePath = self.makeDir()

    def createHubForInstallService(self, downloadFilePaths=None,
                                   eucaImageRegistration=None,
                                   storageStoredImages=None,
                                   storageStoredImageStates=None,
                                   storagePreImageRegistration=None):
        if storageStoredImages is None:
            storageStoredImages = []
        if storageStoredImageStates is None:
            storageStoredImageStates = []

        serviceHub = ServiceHub()

        downloadService = FakeDownloadService(
            downloadFilePaths=downloadFilePaths)
        eucaService = FakeEucaService(
            imageRegistration=eucaImageRegistration)
        storageService = FakeStorageService(
            storedImages=storageStoredImages,
            storedImageStates=storageStoredImageStates,
            preImageRegistration=storagePreImageRegistration)
        installService = InstallService(serviceHub)

        serviceHub.addService(installService)
        serviceHub.addService(downloadService)
        serviceHub.addService(eucaService)
        serviceHub.addService(storageService)


        return serviceHub

    def getImageFilesByKind(self, image):
        byKind = {}

        for imageFile in image["files"]:
            kind = imageFile["kind"]
            byKind[kind] = imageFile
        return byKind

    def createImageRegistration(self, image):
        byKind = self.getImageFilesByKind(image)

        imageRegistration = ImageRegistration(image)

        eki = imageRegistration.eki
        eki.path = os.path.join(
            self.basePath, quote(byKind["kernel"]["url"], safe=""))
        eki.sha256 = byKind["kernel"]["sha256"]
        eki.size = byKind["kernel"]["size-in-bytes"]
        eki.kind = "kernel"

        eri = imageRegistration.eri
        eri.path = os.path.join(
            self.basePath, quote(byKind["ramdisk"]["url"], safe=""))
        eri.sha256 = byKind["ramdisk"]["sha256"]
        eri.size = byKind["ramdisk"]["size-in-bytes"]
        eri.kind = "ramdisk"

        emi = imageRegistration.emi
        emi.path = os.path.join(
            self.basePath, quote(byKind["image"]["url"], safe=""))
        emi.sha256 = byKind["image"]["sha256"]
        emi.size = byKind["image"]["size-in-bytes"]
        emi.kind = "image"
        return imageRegistration

    def getDownloadFilePaths(self, imageRegistration):
        byKind = self.getImageFilesByKind(imageRegistration.image)

        downloadPaths = {}
        downloadPaths[byKind["kernel"]["url"]] = imageRegistration.eki.path
        downloadPaths[byKind["ramdisk"]["url"]] = imageRegistration.eri.path
        downloadPaths[byKind["image"]["url"]] = imageRegistration.emi.path
        return downloadPaths

    def testDownloadImage(self):
        image1 = self.createImage(1, withFiles=True)
        expectedImageRegistration = self.createImageRegistration(image1)

        def callback(result, serviceHub):
            self.assertEquals(result, expectedImageRegistration)

            eucaService = serviceHub.getService(FakeEucaService)
            self.assertEquals(len(eucaService.tasks), 1)
            self.assertIdentical(type(eucaService.tasks[0]),
                                 BundleAndUploadImageTask)

            task = eucaService.tasks[0]
            self.assertEquals(task.imageRegistration, expectedImageRegistration)

            storageService = serviceHub.getService(FakeStorageService)
            self.assertEquals(len(storageService.tasks), 1)
            self.assertIdentical(type(storageService.tasks[0]),
                                 SetImageRegistrationTask)

            task = storageService.tasks[0]
            self.assertEquals(task.imageRegistration,
                              eucaService.tasks[0].imageRegistration)

        downloadFilePaths = self.getDownloadFilePaths(expectedImageRegistration)
        serviceHub = self.createHubForInstallService(
            downloadFilePaths=downloadFilePaths,
            eucaImageRegistration=expectedImageRegistration)

        installService = serviceHub.getService(InstallService)
        deferred = installService.addTask(InstallImageTask(image1))
        deferred.addCallback(callback, serviceHub)

        return self.runServicesAndWaitForDeferred(
            [serviceHub], deferred)

    def testGetStates(self):
        image1 = self.createImage(1, withFiles=True)
        byKind = self.getImageFilesByKind(image1)
        expectedImageRegistration = self.createImageRegistration(image1)

        downloadFilePaths = self.getDownloadFilePaths(expectedImageRegistration)
        kernelUrl = byKind["kernel"]["url"]
        kernelFilePath = downloadFilePaths[kernelUrl]
        downloadFilePaths[kernelUrl] = kernelDeferred = Deferred()

        def storedImageState(imageRegistration):
            if imageRegistration is None:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "uninstalled"})]
            else:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "installed"})]

        eucaBundleDeferred = Deferred()
        serviceHub = self.createHubForInstallService(
            downloadFilePaths=downloadFilePaths,
            eucaImageRegistration=eucaBundleDeferred,
            storageStoredImageStates=storedImageState)

        installService = serviceHub.getService(InstallService)

        def installImagesCallback(result):
            self.assertEquals(result, expectedImageRegistration)

        def checkInstalledCallback(result):
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(installedCallback)
            return deferred

        def installedCallback(states):
            self.assertEquals(states[0]["status"], "installed")

        installDeferred = installService.addTask(InstallImageTask(image1))
        installDeferred.addCallback(installImagesCallback)
        installDeferred.addCallback(checkInstalledCallback)

        def checkDownloadingCallback(states):
            self.assertEquals(len(states), 1)
            self.assertEquals(states[0]["status"], "downloading")

        def installingCallback(states):
            self.assertEquals(states[0]["status"], "installing")
            eucaBundleDeferred.callback(expectedImageRegistration)

        def checkInstallingCallback(result):
            kernelDeferred.callback(kernelFilePath)
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(installingCallback)
            return deferred

        statusDeferred = installService.addTask(GetStatesTask([image1["uri"]]))
        statusDeferred.addCallback(checkDownloadingCallback)
        statusDeferred.addCallback(checkInstallingCallback)

        return self.runServicesAndWaitForDeferred(
            [serviceHub], mergeDeferreds([statusDeferred, installDeferred]))

    def testDownloadFailedRevertsStatusToUninstalled(self):
        image1 = self.createImage(1, withFiles=True)
        byKind = self.getImageFilesByKind(image1)
        expectedImageRegistration = self.createImageRegistration(image1)

        downloadFilePaths = self.getDownloadFilePaths(expectedImageRegistration)
        kernelUrl = byKind["kernel"]["url"]
        kernelFilePath = downloadFilePaths[kernelUrl]
        downloadFilePaths[kernelUrl] = kernelDeferred = Deferred()

        def storedImageState(imageRegistration):
            if imageRegistration is None:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "uninstalled"})]
            else:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "installed"})]

        serviceHub = self.createHubForInstallService(
            downloadFilePaths=downloadFilePaths,
            storageStoredImageStates=storedImageState)

        installService = serviceHub.getService(InstallService)

        def installImagesCallback(result):
            self.fail("Should not succeed since the download failed.")

        def installImagesErrback(failure):
            failure.trap(DownloadServiceError)
            self.assertEquals(failure.value.message, "!ERROR!")

        def checkUninstalledCallback(result):
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(uninstalledCallback)
            return deferred

        def uninstalledCallback(states):
            self.assertEquals(states[0]["status"], "uninstalled")

        installDeferred = installService.addTask(InstallImageTask(image1))
        installDeferred.addCallbacks(installImagesCallback, installImagesErrback)
        installDeferred.addCallback(checkUninstalledCallback)

        def checkDownloadingCallback(states):
            self.assertEquals(len(states), 1)
            self.assertEquals(states[0]["status"], "downloading")
            kernelDeferred.errback(DownloadServiceError("!ERROR!"))

        statusDeferred = installService.addTask(GetStatesTask([image1["uri"]]))
        statusDeferred.addCallback(checkDownloadingCallback)

        return self.runServicesAndWaitForDeferred(
            [serviceHub], installDeferred)

    def testBundleFailedRevertsStatusToUninstalled(self):
        image1 = self.createImage(1, withFiles=True)
        byKind = self.getImageFilesByKind(image1)
        expectedImageRegistration = self.createImageRegistration(image1)

        downloadFilePaths = self.getDownloadFilePaths(expectedImageRegistration)
        kernelUrl = byKind["kernel"]["url"]
        kernelFilePath = downloadFilePaths[kernelUrl]
        downloadFilePaths[kernelUrl] = kernelDeferred = Deferred()

        def storedImageState(imageRegistration):
            if imageRegistration is None:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "uninstalled"})]
            else:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "installed"})]

        eucaBundleDeferred = Deferred()
        serviceHub = self.createHubForInstallService(
            downloadFilePaths=downloadFilePaths,
            eucaImageRegistration=eucaBundleDeferred,
            storageStoredImageStates=storedImageState)

        installService = serviceHub.getService(InstallService)

        def installImagesCallback(result):
            self.fail("Should not succeed since the bundle failed.")

        def installImagesErrback(failure):
            failure.trap(EucaServiceError)
            self.assertEquals(failure.value.message, "!ERROR!")

        def checkUninstalledCallback(result):
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(uninstalledCallback)
            return deferred

        def uninstalledCallback(states):
            self.assertEquals(states[0]["status"], "uninstalled")

        installDeferred = installService.addTask(InstallImageTask(image1))
        installDeferred.addCallbacks(installImagesCallback, installImagesErrback)
        installDeferred.addCallbacks(checkUninstalledCallback)

        def checkDownloadingCallback(states):
            self.assertEquals(len(states), 1)
            self.assertEquals(states[0]["status"], "downloading")

        def installingCallback(states):
            self.assertEquals(states[0]["status"], "installing")
            eucaBundleDeferred.errback(EucaServiceError("!ERROR!"))

        def checkInstallingCallback(result):
            kernelDeferred.callback(kernelFilePath)
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(installingCallback)
            return deferred

        statusDeferred = installService.addTask(GetStatesTask([image1["uri"]]))
        statusDeferred.addCallback(checkDownloadingCallback)
        statusDeferred.addCallback(checkInstallingCallback)

        return self.runServicesAndWaitForDeferred(
            [serviceHub], mergeDeferreds([statusDeferred, installDeferred]))

    def testSetImageRegistrationFailedRevertsStatusToUninstalled(self):
        image1 = self.createImage(1, withFiles=True)
        byKind = self.getImageFilesByKind(image1)
        expectedImageRegistration = self.createImageRegistration(image1)

        downloadFilePaths = self.getDownloadFilePaths(expectedImageRegistration)
        kernelUrl = byKind["kernel"]["url"]
        kernelFilePath = downloadFilePaths[kernelUrl]
        downloadFilePaths[kernelUrl] = kernelDeferred = Deferred()

        def storedImageState(imageRegistration):
            if imageRegistration is None:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "uninstalled"})]
            else:
                return [ImageState({"image-uri": image1["uri"],
                                    "status": "installed"})]

        def preImageRegistrationFails():
            raise StorageServiceError("!ERROR!")

        eucaBundleDeferred = Deferred()
        serviceHub = self.createHubForInstallService(
            downloadFilePaths=downloadFilePaths,
            eucaImageRegistration=eucaBundleDeferred,
            storageStoredImageStates=storedImageState,
            storagePreImageRegistration=preImageRegistrationFails)

        installService = serviceHub.getService(InstallService)

        def installImagesCallback(result):
            self.fail("Should not succeed since saving image registration failed.")

        def installImagesErrback(failure):
            failure.trap(StorageServiceError)
            self.assertEquals(failure.value.message, "!ERROR!")

        def checkUninstalledCallback(result):
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(uninstalledCallback)
            return deferred

        def uninstalledCallback(states):
            self.assertEquals(states[0]["status"], "uninstalled")

        installDeferred = installService.addTask(InstallImageTask(image1))
        installDeferred.addCallbacks(installImagesCallback, installImagesErrback)
        installDeferred.addCallbacks(checkUninstalledCallback)

        def checkDownloadingCallback(states):
            self.assertEquals(len(states), 1)
            self.assertEquals(states[0]["status"], "downloading")

        def installingCallback(states):
            self.assertEquals(states[0]["status"], "installing")
            eucaBundleDeferred.callback(expectedImageRegistration)

        def checkInstallingCallback(result):
            kernelDeferred.callback(kernelFilePath)
            deferred = installService.addTask(GetStatesTask([image1["uri"]]))
            deferred.addCallback(installingCallback)
            return deferred

        statusDeferred = installService.addTask(GetStatesTask([image1["uri"]]))
        statusDeferred.addCallback(checkDownloadingCallback)
        statusDeferred.addCallback(checkInstallingCallback)

        return self.runServicesAndWaitForDeferred(
            [serviceHub], mergeDeferreds([statusDeferred, installDeferred]))
